"""
Tests for command line interface (CLI)
"""
import os
import pytest
import re
import warnings
import primalscheme.cli

from argparse import Namespace
from Bio.SeqRecord import SeqRecord

with warnings.catch_warnings():
    warnings.simplefilter("ignore", PendingDeprecationWarning)
    from primalscheme.cli import process_fasta


def test_runas_module():
    """
    Can this package be run as a Python module?
    """
    exit_status = os.system("python -m primalscheme --help")
    assert exit_status == 0


def test_entrypoint():
    """
    Is entrypoint script installed? (setup.py)
    """
    exit_status = os.system("primalscheme --help")
    assert exit_status == 0


def test_mutliplex_command():
    """
    Is multiplex command available?
    """
    exit_status = os.system("primalscheme multiplex --help")
    assert exit_status == 0


def test_cli_fails_without_command():
    """
    Does CLI stop execution w/o a command argument?
    """
    with pytest.raises(SystemExit):
        primalscheme.cli.main()
        pytest.fail("CLI doesn't abort asking for a command argument")


def test_cli_fails_without_fasta():
    """
    Does CLI stop execution w/o a positional fasta arg?
    """
    with pytest.raises(SystemExit):
        args = ["multiplex"]
        primalscheme.cli.parse_arguments(args)


@pytest.mark.parametrize(
    "option, value",
    [
        ("--prefix", "myScheme"),
        ("--amplicon-size-min", "380"),
        ("--amplicon-size-max", "420"),
        ("--target-overlap", "0"),
        ("--output-path", "./scheme"),
        ("--step-distance", "11"),
        ("--debug", None),
        ("--no-sort", None),
        ("--force", None),
    ],
)
def test_availability_of_cli_options(option, value):
    """
    Are options available?
    """
    args = ["multiplex", "some.fa", option]
    if value:
        args.append(value)
    parsed = primalscheme.cli.parse_arguments(args)
    assert isinstance(parsed, Namespace)


def test_cli_fails_with_invalid_option(chikv_input):
    """
    Does CLI stop execution with an invalid option?
    """
    with pytest.raises(SystemExit):
        args = ["multiplex", f"{str(chikv_input)}", "--force", "--badOption"]
        primalscheme.cli.parse_arguments(args)


@pytest.mark.parametrize(
    "option", ["-V", "--version"],
)
def test_cli_version_output(option, capsys):
    """
    Does CLI output a sensible version number for -V and --version?
    """
    try:
        primalscheme.cli.parse_arguments([option])
    except SystemExit:
        pass
    out, err = capsys.readouterr()
    assert re.match(r"^primalscheme \d{1,2}.\d{1,2}.\d{1,2}[a-zA-Z]{0,3}\d*$", out)


def test_force_not_required_when_output_path_does_not_exist(tmp_path):
    path = tmp_path / "output"
    primalscheme.cli.get_output_path(output_path=path)


def test_force_required_when_output_path_does_exist(tmp_path):
    path = tmp_path / "output"
    path.mkdir()
    with pytest.raises(IOError):
        primalscheme.cli.get_output_path(output_path=path)


def test_force_allows_existing_output_path(tmp_path):
    path = tmp_path / "output"
    path.mkdir()
    primalscheme.cli.get_output_path(output_path=path, force=True)


def test_existing_output_path_not_dir(tmp_path):
    path = tmp_path / "output"
    path.write_text("text")
    with pytest.raises(IOError):
        primalscheme.cli.get_output_path(output_path=path, force=True)


def test_too_short_fasta_size_vs_amplicon_size(input_fasta_short_500, tmp_path, caplog):
    args = ["multiplex", str(input_fasta_short_500)]
    parsed = primalscheme.cli.parse_arguments(args)
    outpath = tmp_path / "output"
    output_path = primalscheme.cli.get_output_path(outpath)

    with pytest.raises(SystemExit):
        primalscheme.cli.multiplex(parsed, output_path)
    messages = [log.message for log in caplog.records]
    assert any("too short" in msg for msg in messages)


def test_process_fasta_empty_input(input_fasta_empty):
    with pytest.raises(ValueError, match="does not contain any valid references"):
        process_fasta(input_fasta_empty)


def test_process_fasta_invalid_alphabet(input_fasta_invalid_alphabet):
    with pytest.raises(ValueError, match="invalid nucleotide codes"):
        process_fasta(input_fasta_invalid_alphabet)


def test_process_fasta_too_many_records(input_fasta_101_random_valid):
    with pytest.raises(ValueError, match="A maximum of 100"):
        process_fasta(input_fasta_101_random_valid)


def test_process_fasta_size_difference_over_500(input_fasta_size_difference_over_500,):
    with pytest.raises(ValueError, match="too different in size"):
        process_fasta(input_fasta_size_difference_over_500)


def test_process_fasta_returns_list_of_seq_records(input_fasta_5_random_valid):
    references = process_fasta(input_fasta_5_random_valid)
    assert len(references) == 5
    assert all(isinstance(r, SeqRecord) for r in references)


def test_process_fasta_chikv_demo(chikv_input):
    references = process_fasta(chikv_input)
    assert len(references) == 2


def test_process_fasta_gaps_removed(input_fasta_valid_with_gaps):
    references = process_fasta(input_fasta_valid_with_gaps)
    assert "-" not in references[0]


def test_process_fasta_too_short_input(input_fasta_short_500):
    with pytest.raises(ValueError, match="too short"):
        process_fasta(input_fasta_short_500, min_ref_size=900)


def test_process_fasta_returns_longest_reference_first(input_fasta_shortest_first):
    references = process_fasta(input_fasta_shortest_first, sort=True)
    first_ref_len = len(references[0])
    for ref in references[1:]:
        assert len(ref) <= first_ref_len


def test_process_fasta_no_sort(input_fasta_shortest_first):
    references = process_fasta(input_fasta_shortest_first, sort=False)
    first_ref_len = len(references[0])
    for ref in references[1:]:
        assert len(ref) >= first_ref_len



"""
PrimalScheme: a primer3 wrapper for designing multiplex primer schemes

Copyright (C) 2020 Joshua Quick and Andrew Smith
www.github.com/aresti/primalscheme

This module contains the CLI for PrimalScheme.
It is executed when the user runs 'primalscheme' after installation,
or 'python -m primalscheme'.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>
"""

import argparse
import logging
import sys

from pathlib import Path

from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from progress.bar import ShadyBar

from primalscheme import __version__ as version, config
from primalscheme.region import NoSuitablePrimersError
from primalscheme.reporting import MultiplexReporter, ProgressTracker

logger = logging.getLogger("primalscheme")


def main():
    """
    Entry point for primalscheme.
    Process args, set up logging and call the command function.
    """

    # Parse args, update config
    try:
        args = parse_arguments(sys.argv[1:])
    except ValueError as e:
        print(f"Error: {e}")
        sys.exit(1)

    # Validate output path
    try:
        outpath = get_output_path(args.output_path, force=args.force)
    except IOError as e:
        print(f"Error: {e}")
        sys.exit(1)

    # Setup logging
    setup_logging(outpath, debug=args.debug, prefix=args.prefix)

    for arg in vars(args):
        logger.debug("{}: {}".format(arg, str(vars(args)[arg])))

    # Run subcommand
    args.func(args, outpath)


def multiplex(args, outpath):
    """Multipex scheme command."""

    # Process FASTA input
    min_ref_size = args.amplicon_size_max * 2
    sort = not args.no_sort
    try:
        references = process_fasta(args.fasta, min_ref_size=min_ref_size, sort=sort)
    except ValueError as e:
        logger.error(f"Error: {e}")
        sys.exit(2)

    # Progress bar
    progress_bar = ProgressBar()

    # Create scheme
    try:
        scheme = MultiplexReporter(
            outpath,
            references,
            prefix=args.prefix,
            amplicon_size_min=args.amplicon_size_min,
            amplicon_size_max=args.amplicon_size_max,
            target_overlap=args.target_overlap,
            progress_tracker=progress_bar,
        )
        scheme.design_scheme()
    except NoSuitablePrimersError:
        # Unable to find primers for at least one region
        logger.error("Error: Unable to find suitable primers")
        sys.exit(10)
    except Exception as e:
        # Unexpected error
        logger.error(f"Error: {e}")
        raise e

    # Write outputs
    logger.info(
        f"All done! Scheme created with {len(scheme.regions)} regions, "
        f"{scheme.gap_count } gap{'' if scheme.gap_count == 1 else 's'}, "
        f"{scheme.percent_coverage}% coverage"
    )
    scheme.write_default_outputs()
    if args.debug:
        scheme.write_pickle()
    sys.exit(0)


def process_fasta(file_path, min_ref_size=None, sort=True):
    """Parse and validate the fasta file."""

    references = []
    records = SeqIO.parse(file_path, "fasta")  # may raise

    # Remove gaps
    for record in records:
        ref = SeqRecord(
            Seq(str(record.seq).replace("-", "").upper()),
            id=record.id,
            description=record.id,
        )
        references.append(ref)

    # Check for no references
    if not references:
        raise ValueError("The input FASTA file does not contain any valid references.")

    # Check for too short references
    if min_ref_size and any(len(ref) < min_ref_size for ref in references):
        raise ValueError(
            "One or more of your references is too short. Based on your target "
            f"amplicon size, the minimum reference size is {min_ref_size} nt."
        )

    # Check for too many references
    if len(references) > 100:
        raise ValueError("A maximum of 100 reference genomes is currently supported.")

    # Check for max difference in size between references
    primary_ref = references[0]
    primary_ref_len = len(primary_ref)

    if any(abs(len(r) - primary_ref_len) > 500 for r in references):
        raise ValueError(
            "One or more of your references is too different in size to "
            "the primary (first) reference. The maximum difference is "
            "500 nt."
        )

    # Check for a valid alphabet
    VALID_ALPHABET = "AGCTacgt"
    for r in references:
        if any(base not in VALID_ALPHABET for base in set(r.seq)):
            raise ValueError(
                "One or more of your fasta sequences contain invalid "
                f"nucleotide codes. The supported alphabet is '{VALID_ALPHABET}'. "
                "Ambiguity codes and gaps are not currently supported."
            )

    # Sort references by length to maximum scheme coverage
    if sort:
        references.sort(key=len, reverse=True)
    return references


def setup_logging(output_path, debug=False, prefix="primalscheme"):
    """Setup logging output and verbosity."""

    logger.setLevel(logging.DEBUG if debug else logging.INFO)

    # File handler
    log_filepath = output_path / f"{prefix}.log"
    fh = logging.FileHandler(log_filepath)
    fh.setLevel(logging.DEBUG)
    fh_formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    fh.setFormatter(fh_formatter)
    logger.addHandler(fh)

    # Stream handler STDOUT
    sh = logging.StreamHandler(sys.stdout)
    sh.setLevel(logging.INFO)
    sh_formatter = logging.Formatter("%(message)s")
    sh.setFormatter(sh_formatter)
    logger.addHandler(sh)

    logger.info(f"Writing log to {log_filepath}")


def get_output_path(output_path, force=False):
    """
    Check for an existing output dir, require --force to overwrite.
    Create dir, return path object.
    """
    path = Path(output_path)

    if path.exists() and not force:
        raise IOError("Directory exists add --force to overwrite")
    elif path.exists() and not path.is_dir():
        raise IOError("The output path is not a directory.")

    path.mkdir(exist_ok=True)
    return path


def parse_arguments(args):
    """Parse command line arguments."""
    # Setup parsers
    parser = argparse.ArgumentParser(
        prog="primalscheme",
        description=("A primer3 wrapper for designing multiplex primer schemes."),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    subparsers = parser.add_subparsers(title="[subcommands]", dest="command")
    subparsers.required = True

    parser_scheme = subparsers.add_parser("multiplex", help="Multiplex PCR scheme")

    # Set func
    parser_scheme.set_defaults(func=multiplex)

    # Add arguments to parser
    parser_scheme.add_argument("fasta", help="FASTA file")
    parser_scheme.add_argument(
        "--prefix",
        default=config.PREFIX,
        help="Prefix used for primer names and output files " "(default: %(default)s)",
    )
    parser_scheme.add_argument(
        "--amplicon-size-min",
        type=positive_int,
        default=config.AMPLICON_SIZE_MIN,
        help="Minimum amplicon size (default: %(default)i)",
    )
    parser_scheme.add_argument(
        "--amplicon-size-max",
        type=positive_int,
        default=config.AMPLICON_SIZE_MAX,
        help="Maximum amplicon size (default: %(default)i)",
    )
    parser_scheme.add_argument(
        "--output-path",
        default=config.OUTPUT_PATH,
        help="Output directory (default: %(default)s)",
    )
    parser_scheme.add_argument(
        "--target-overlap",
        default=config.TARGET_OVERLAP,
        type=positive_int,
        help="Target overlap size (default: %(default)i)",
    )
    parser_scheme.add_argument(
        "--step-distance",
        default=config.STEP_DISTANCE,
        type=positive_int,
        help="Distance to step between find attempts (default: %(default)i)",
    )
    parser_scheme.add_argument("--debug", action="store_true", help="Verbose logging")
    parser_scheme.add_argument(
        "--no-sort",
        action="store_true",
        help="Don't sort input FASTA by length (will use first reference in BED file",
    )
    parser_scheme.add_argument(
        "--force",
        action="store_true",
        help="Force output to an existing directory and overwrite output " "files",
    )
    parser.add_argument(
        "-V", "--version", action="version", version=f"%(prog)s {version}"
    )

    # Generate args
    parsed = parser.parse_args(args)

    # Post parsing checks
    if parsed.amplicon_size_max < parsed.amplicon_size_min:
        raise ValueError("--amplicon-size-min cannot exceed --amplicon-size-max")

    return parsed


def positive_int(string):
    """Validate command line args as being of type: positive int."""
    value = int(string)
    if value < 0:
        raise argparse.ArgumentTypeError("positive integer required.")
    return value


class ProgressBar(ShadyBar, ProgressTracker):
    """Progress bar for terminal stdout."""

    suffix = "%(percent)d%% [%(index)d / %(max)d]"

    def __init__(self, *args, **kwargs):
        """Init ProgressBar."""
        self.__considered = 0
        self.__region_num = 1
        super().__init__(*args, **kwargs)

    def goto(self, val):
        """Update progress to val."""
        super().goto(val)

    @property
    def end(self):
        """End (max) progress value."""
        return self.max

    @end.setter
    def end(self, val):
        """Set end progress value."""
        self.max = val

    @property
    def considered(self):
        """Count of considered primers."""
        return self.__considered

    @considered.setter
    def considered(self, considered):
        """Set count of considered primers."""
        self.__considered = considered
        self.update_message()

    @property
    def region_num(self):
        """Current region num."""
        return self.__region_num

    @region_num.setter
    def region_num(self, region_num):
        """Set current region num."""
        self.__region_num = region_num
        self.update_message()

    def update_message(self):
        """Update progress bar prefix message."""
        self.message = f"Considered {self.considered} primers, region {self.region_num}"

    def interrupt(self):
        """Prepare to be interrupted by a log message."""
        if self.index:
            self.finish()


if __name__ == "__main__":
    main()
